#include <QtGui>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <ctime>
#include "bsdetectionwidget.h"

using namespace std;


const int BSDetectionWidget::DEFAULT_PEN_WIDTH = 1;



BSDetectionWidget::BSDetectionWidget (QWidget *parent)
{
  Q_UNUSED (parent);

  // Sets default user interface parameters
  setFocus ();
  grabKeyboard ();
  udef = false;

  // Sets initial values for the gradient map
  gMap = NULL;

  // Initializes the auxiliary views
  accuview = NULL;
  strucview = NULL;
  profileview = NULL;

  alternate = 0;
  verbose = false;
  paintItBlack = false;
  exam = -1;

  selectionColor = Qt::red;
  bsColor = Qt::yellow;
  bsHighColor = Qt::blue;
  bsPointsVisible = true;
  boundColor = Qt::magenta;
  boundHighColor = Qt::green;
  bsBoundsVisible = false;
}


BSDetectionWidget::~BSDetectionWidget ()
{
  if (accuview != NULL) delete accuview;
  if (strucview != NULL) delete strucview;
}


QSize BSDetectionWidget::openImage (const QString &fileName, int type)
{
  QSize newSize (0, 0);
  loadedImage.load (fileName);
  width = loadedImage.width ();
  height = loadedImage.height ();
  newSize = loadedImage.size ();
 
  augmentedImage = loadedImage;
  if (gMap != NULL) delete gMap;
  gMap = new VMap (width, height, getBitmap (augmentedImage), type);
  detector.setGradientMap (gMap);
 
  update ();
  if (profileview != NULL) profileview->setImage (&loadedImage, gMap);
  if (strucview != NULL) strucview->setGradientMap (gMap);

  return newSize;
}


int **BSDetectionWidget::getBitmap (const QImage &image)
{
  int w = image.width ();
  int h = image.height ();
  
  int **tabImage = new int*[h];
  for (int i = 0; i < h; i++)
  {
    tabImage[i] = new int[w];
    for(int j = 0; j < w; j++)
    {
      QColor c = QColor (image.pixel (j, h - i - 1));
      tabImage[i][j] = c.value ();
    }
  }
  return tabImage;
}


bool BSDetectionWidget::saveAugmentedImage (const QString &fileName,
                                            const char *fileFormat)
{
  QImage aImage = augmentedImage;
  return (aImage.save (fileName, fileFormat));
}


void BSDetectionWidget::clearImage ()
{
  augmentedImage.fill (qRgb (255, 255, 255));
  update ();
}


void BSDetectionWidget::paintEvent (QPaintEvent *)
{
  QPainter painter (this);
  painter.drawImage (QPoint (0, 0), augmentedImage);
}


void BSDetectionWidget::closeAccuAnalyzer ()
{
  if (accuview != NULL)
  {
    accuview->close ();
    delete accuview;
    accuview = NULL;
  }
}


void BSDetectionWidget::closePixelAnalyzer ()
{
  if (strucview != NULL)
  {
    strucview->close ();
    delete strucview;
    strucview = NULL;
  }
}


void BSDetectionWidget::closeProfileAnalyzer ()
{
  if (profileview != NULL)
  {
    profileview->close ();
    delete profileview;
    profileview = NULL;
  }
}


void BSDetectionWidget::switchAccuAnalyzer ()
{
  if (accuview != NULL)
  {
    accuview->close ();
    delete accuview;
    accuview = NULL;
  }
  else
  {
    accuview = new BSAccumulatorView (&detector);
    accuview->show ();
  }
}


void BSDetectionWidget::switchPixelAnalyzer ()
{
  if (strucview != NULL)
  {
    strucview->close ();
    delete strucview;
    strucview = NULL;
  }
  else
  {
    strucview = new BSStructureView (&loadedImage, &detector);
    strucview->setGradientMap (gMap);
    strucview->show ();
  }
}


void BSDetectionWidget::switchProfileAnalyzer ()
{
  if (profileview != NULL)
  {
    profileview->close ();
    delete profileview;
    profileview = NULL;
  }
  else
  {
    profileview = new BSProfileView ();
    profileview->setImage (&loadedImage, gMap);
    if (! p1.equals (p2)) profileview->buildScans (p1, p2);
    profileview->show ();
  }
}


void BSDetectionWidget::mousePressEvent (QMouseEvent *event)
{
  this->p1 = Pt2i (event->pos().x (), height - 1 - event->pos().y());
  udef = true;
}


void BSDetectionWidget::mouseReleaseEvent (QMouseEvent *event)
{
  this->p2 = Pt2i (event->pos().x (), height - 1 - event->pos().y());
alternate = 0;
  cerr << "p1 defined: " << p1.x () << " " << p1.y () << endl;
  cerr << "p2 defined: " << p2.x () << " " << p2.y () << endl;
  extract ();
}


void BSDetectionWidget::mouseMoveEvent (QMouseEvent *event)
{
  this->p2 = Pt2i (event->pos().x (), height - 1 - event->pos().y ());
  if (verbose) cerr << "(" << p1.x () << ", " << p1.y () << ") ("
                    << p2.x () << ", " << p2.y () << ")" << endl;
  if (p1.manhattan (p2) > 5
      && (width > p2.x() && height > p2.y()
          && p2.x() > 0 && p2.y() > 0))
  {
    extract (false);
  }
}


void BSDetectionWidget::keyPressEvent (QKeyEvent *event)
{
  if (isActiveWindow ()) switch (event->key ())
  {
    case Qt::Key_U :
      cerr << "p1 update: " << p1.x () << " " << p1.y () << endl;
      cerr << "p2 update: " << p2.x () << " " << p2.y () << endl;
      extract ();
      break;

    case Qt::Key_A :
      if (event->modifiers () & Qt::ShiftModifier)
      {
        bsBoundsVisible = ! bsBoundsVisible;
        cout << "Segments bounds "
             << (bsBoundsVisible ? "visible" : "hidden") << endl;
      }
      else
      {
        bsPointsVisible = ! bsPointsVisible;
        cout << "Segments points "
             << (bsPointsVisible ? "visible" : "hidden") << endl;
      }
      break;

    case Qt::Key_M : // Multi-selection switch
      if (event->modifiers () & Qt::ShiftModifier)
      {
        detector.switchMultiSelection ();
        cout << "Multi-selection "
             << (detector.isMultiSelection () ? "on" : "off") << endl;
      }
      else
      {
        if (event->modifiers () & Qt::ControlModifier)
          paintItBlack = ! paintItBlack;
        extractAll ();
      }
      break;

    case Qt::Key_L : // Next multi-selected segment
      if (! detector.getBlurredSegments().empty ())
      {
        displayNext (event->modifiers () & Qt::ShiftModifier ? -1 : 1);
        cout << "Selection du segment " << exam << endl;
      }
      break;

    case Qt::Key_Y : // Initial detection extension limitation
      detector.switchInitialBounding ();
      extract ();
      cout << "Initial step max extension = "
           << detector.initialDetectionMaxExtent () << endl;
      break;

    case Qt::Key_D : // Density test at initial step
      detector.switchDensityTest ();
      extract ();
      cout << "Density test : "
           << (detector.isSetDensityTest () ? "on" : "off") << endl;
      break;

    case Qt::Key_W : // Fast tracks max width
      if (event->modifiers () & Qt::ControlModifier)
      {
        detector.switchScanRecentering ();
        cout << "Fine tracking centered on " << (detector.isScanRecentering () ?
                "detected segment" : "initial scan") << endl;
        extract ();
      }
      else
      {
        detector.setFineTracksMaxWidth (detector.fineTracksMaxWidth () +
          (event->modifiers () & Qt::ShiftModifier ? -1 : 1));
        extract ();
        cout << "Fine tracks max width = "
             << detector.fineTracksMaxWidth () << endl;
      }
      break;

    case Qt::Key_X :
      if (event->modifiers () & Qt::ControlModifier)
      {
        detector.switchScanFitting ();
        cout << "Fine tracking fitted to " << (detector.isScanFitting () ?
                "detected segment width" : "assigned width") << endl;
        extract ();
      }
      else
      {
        detector.setFastTracksMaxWidth (detector.fastTracksMaxWidth () +
          (event->modifiers () & Qt::ShiftModifier ? -1 : 1));
        extract ();
        cout << "Fast tracks max width = "
             << detector.fastTracksMaxWidth () << endl;
      }
      break;

    case Qt::Key_K : // Output blurred segment min size
      detector.setBSminSize (detector.getBSminSize () +
        (event->modifiers () & Qt::ShiftModifier ? -1 : 1));
      extract ();
      cout << "Output BS min size = " << detector.getBSminSize () << endl;
      break;

    case Qt::Key_H : // Pixel lack tolerence
      detector.setPixelLackTolerence (detector.getPixelLackTolerence () +
        (event->modifiers () & Qt::ShiftModifier ? -1 : 1));
      extract ();
      cout << "Lack tolerence = " << detector.getPixelLackTolerence ()
           << " pixels" << endl;
      break;

    case Qt::Key_N : // Automatic detection grid resolution
      detector.setAutoGridResolution (detector.getAutoGridResolution () +
        (event->modifiers () & Qt::ShiftModifier ? -1 : 1));
      cout << "Auto grid resolution = "
           << detector.getAutoGridResolution () << " pixels" << endl;
      break;

    case Qt::Key_Q :
      detector.switchAutoRestart ();
      extract ();
      cout << "Segment continuation after = "
           << detector.getRestartOnLack () << " pixels" << endl;
      break;

    case Qt::Key_T :
      detector.toggleThinning ();
      extract ();
      cout << "Thinning "
           << (detector.isThinningActivated () ? "on" : "off") << endl;
      break;

    case Qt::Key_S :
      detector.switchDynamicScans ();
      cout << (detector.dynamicScansOn () ?
               "Dynamic scans" : "Static scans") << endl;
      extract ();
      break;

    case Qt::Key_O :
      detector.switchOrthoScans ();
      cout << (detector.orthoScansOn () ?
               "Orthographic scans" : "Directional scans") << endl;
      extract ();
      break;

    case Qt::Key_F :
      detector.switchFiltering (1);
      cout << "Final filtering "
           << (detector.isFiltering (1) ? "on" : "off") << endl;
      extract ();
      break;

    case Qt::Key_P :
      if (event->modifiers () & Qt::ControlModifier)
      {
        writeDetectionResult ();
        cout << "Detection result output" << endl;
      }
      else if (event->modifiers () & Qt::ShiftModifier)
      {
        augmentedImage.save ("capture.png");
        cout << "Main window shot in capture.png" << endl;
      }
      else
      {
        detector.switchFiltering (0);
        cout << "Pre-filtering "
             << (detector.isFiltering (0) ? "on" : "off") << endl;
        extract ();
      }
      break;

    case Qt::Key_G :
      detector.switchConnectivityConstraint ();
      cout << "Connectivity constraint "
           << (detector.isSetConnectivityConstraint () ? "on" : "off") << endl;
      extract ();
      break;

    case Qt::Key_C :
      if (event->modifiers () & Qt::ControlModifier)
        saveExtractedSegment ();
      break;

    case Qt::Key_V :
      if (event->modifiers () & Qt::ControlModifier)
        displaySavedSegments ();
      else switchVerbose ();
      break;

    case Qt::Key_Z :
      if (event->modifiers () & Qt::ControlModifier)
        clearSavedSegments ();
      break;

    case Qt::Key_E :
      if (event->modifiers () & Qt::ShiftModifier)
        detector.invertEdgeDirection ();
      else detector.switchEdgeDirectionConstraint ();
      switch (detector.edgeDirectionConstraint ())
      {
        case 0 :
          cout << "Stroke detection" << endl;
          break;
        case 1 :
          cout << "Edge detection" << endl;
          break;
        case -1 :
          cout << "Opposite edge detection" << endl;
          break;
      }
      extract ();
      break;

    case Qt::Key_J :
      detector.switchPreliminary ();
      cout << "Preliminary detection "
           << (detector.isPreliminary () ? "on" : "off") << endl;
      extract ();
      break;

    case Qt::Key_1 :
      switchPixelAnalyzer ();
      break;

    case Qt::Key_2 :
      switchAccuAnalyzer ();
      break;

    case Qt::Key_3 :
      switchProfileAnalyzer ();
      break;

    case Qt::Key_8 :
      alternateTest ();
      break;

    case Qt::Key_9 :
      performanceTest ();
      break;

    case Qt::Key_0 :
      localTest ();
      break;
  }
  else if (strucview != NULL && strucview->isActiveWindow ())
  {
    if (strucview->processKeyEvent (event)) extract ();
  }
  else if (accuview != NULL && accuview->isActiveWindow ())
  {
    if (accuview->processKeyEvent (event)) extract ();
  }
  else if (profileview != NULL && profileview->isActiveWindow ())
  {
    if (profileview->processKeyEvent (event)) extract ();
  }
}


void BSDetectionWidget::drawPoints (QPainter &painter,
                                    vector<Pt2i> pts, QColor color)
{
  vector<Pt2i>::iterator iter = pts.begin ();
  while (iter != pts.end ())
  {
    Pt2i p = *iter++;
    painter.setPen (QPen (color, DEFAULT_PEN_WIDTH, Qt::SolidLine,
                          Qt::RoundCap, Qt::RoundJoin));
    if (p.x() < width && p.y() < height && p.x() >= 0 && p.y() >= 0)
      painter.drawPoint (QPoint (p.x(), height - 1 - p.y()));  // dec 1
  }
}


void BSDetectionWidget::drawPixels (QPainter &painter, vector<Pt2i> pix)
{
  vector<Pt2i>::iterator iter = pix.begin ();
  while (iter != pix.end ())
  {
    Pt2i p = *iter++;
    painter.setPen (QPen (QBrush (loadedImage.pixel (p.x (),
                                  loadedImage.height () - 1 - p.y ())),
                          DEFAULT_PEN_WIDTH, Qt::SolidLine,
                          Qt::RoundCap, Qt::RoundJoin));
    if (p.x() < width && p.y() < height && p.x() >= 0 && p.y() >= 0)
      painter.drawPoint (QPoint (p.x(), height - 1 - p.y()));  // dec 1
  }
}


void BSDetectionWidget::drawLine (QPainter &painter,
                                  const Pt2i from, const Pt2i to, QColor color)
{
  int n;
  Pt2i *pts = from.drawing (to, &n);
  painter.setPen (QPen (color, DEFAULT_PEN_WIDTH, Qt::SolidLine,
                        Qt::RoundCap, Qt::RoundJoin));
  for (int i = 0; i < n; i++)
    painter.drawPoint (QPoint (pts[i].x (),
                               height - 1 - pts[i].y ()));  // dec 1
  delete [] pts;
}


void BSDetectionWidget::drawSelection (QPainter &painter,
                                       const Pt2i from, const Pt2i to)
{
  drawLine (painter, from, to, selectionColor);
}


void BSDetectionWidget::drawBlurredSegment (QPainter &painter,
                                            BlurredSegment *bs, bool high)
{
  if (bs != NULL)
  {
    if (bsBoundsVisible)
    {
      vector<Pt2i> bnd;
      DigitalStraightSegment *dss = bs->getSegment ();
      if (dss != NULL)
      {
        dss->getBounds (bnd, 0, 0, width, height);
        drawPoints (painter, bnd, high ? boundHighColor : boundColor);
      }
    }
    if (bsPointsVisible)
      drawPoints (painter, bs->getAllPoints (), high ? bsHighColor : bsColor);
  }
}


void BSDetectionWidget::displayBackground ()
{
  augmentedImage = loadedImage;
  QPainter painter (&augmentedImage);
  update (QRect (QPoint (0, 0), QPoint (width, height)));
}


void BSDetectionWidget::writeDetectionResult ()
{
  BlurredSegment *bs = detector.getBlurredSegment ();
  if (bs != NULL)
  {
    ofstream outf ("seg.txt", ios::out);
    outf << "(" << bs->antipodalEdgeStart().x ()
         << ", " << bs->antipodalEdgeStart().y ()
         << ") (" << bs->antipodalEdgeEnd().x ()
         << ", " << bs->antipodalEdgeEnd().y ()
         << ") (" << bs->antipodalVertex().x ()
         << ", " << bs->antipodalVertex().y ()
         << ") (" << bs->getLastLeft().x ()
         << ", " << bs->getLastLeft().y ()
         << ") (" << bs->getLastRight().x ()
         << ", " << bs->getLastRight().y ()
         << ")" << endl;
    outf.close ();
  }
}


void BSDetectionWidget::displayDetectionResult (bool aux, int hnum)
{
  augmentedImage = loadedImage;
  QPainter painter (&augmentedImage);
  if (paintItBlack) painter.fillRect (0, 0, width, height, QBrush (Qt::black));
  paintItBlack = false;
  vector<BlurredSegment *> bss = detector.getBlurredSegments ();
  if (! bss.empty ())
  {
    int index = 0;
    vector<BlurredSegment *>::const_iterator it = bss.begin ();
    while (it != bss.end ())
      drawBlurredSegment (painter, *it++, index++ == hnum);
  }
  else
    drawBlurredSegment (painter, detector.getBlurredSegment ());
  if (udef) drawSelection (painter, p1, p2);
  update (QRect (QPoint (0, 0), QPoint (width, height)));

  // Update auxiliary view if not dragging
  if (aux)
  {
    if (profileview != NULL)
    {
      profileview->buildScans (p1, p2);
      profileview->scene()->update ();
    }
    if (accuview != NULL) accuview->scene()->update ();
    if (strucview != NULL)
    {
      strucview->setExamined (-1);
      strucview->scene()->update ();
    }
    if (verbose) writeDetectionStatus ();
  }
}


void BSDetectionWidget::displayNext (int dir)
{
  int size = (int) (detector.getBlurredSegments ().size ());
  if (size != 0)
  {
    exam = (exam + dir) % size;
    if (strucview != NULL) strucview->setExamined (exam);
    displayDetectionResult (true, exam);
  }
}


void BSDetectionWidget::displaySavedSegments ()
{
  augmentedImage = loadedImage;
  QPainter painter (&augmentedImage);
  if (! extractedSegments.empty ())
  {
    vector<ExtractedSegment>::iterator it = extractedSegments.begin ();
    while (it != extractedSegments.end ())
    {
      drawBlurredSegment (painter, it->bs);
      drawSelection (painter, it->p1, it->p2);
      it ++;
    }
  }
  update (QRect (QPoint (0, 0), QPoint (width, height)));
}


void BSDetectionWidget::saveExtractedSegment ()
{
  if (detector.isMultiSelection ())
  {
    vector<BlurredSegment *> bss = detector.getBlurredSegments ();
    vector<BlurredSegment *>::const_iterator it = bss.begin ();
    while (it != bss.end ())
    {
      ExtractedSegment es;
      es.bs = *it++;
      es.p1 = p1;
      es.p2 = p2;
      extractedSegments.push_back (es);
    }
    detector.preserveFormerBlurredSegments ();
  }
  else
  {
    BlurredSegment *bs = detector.getBlurredSegment ();
    if (bs != NULL)
    {
      ExtractedSegment es;
      es.bs = bs;
      es.p1 = p1;
      es.p2 = p2;
      extractedSegments.push_back (es);
      detector.preserveFormerBlurredSegment ();
    }
  }
}


void BSDetectionWidget::clearSavedSegments ()
{
  vector<ExtractedSegment>::iterator it = extractedSegments.begin ();
  while (it != extractedSegments.end ())
    delete ((it++)->bs->getSegment ());
  extractedSegments.clear ();
}


void BSDetectionWidget::writeDetectionStatus ()
{
  int res = detector.result ();
  if (res == BSDetector::RESULT_UNDETERMINED)
    cout << "Extraction : undetermined." << endl;
  else if (res == BSDetector::RESULT_OK)
    cout << "Extraction : OK." << endl;
  else if (res == BSDetector::RESULT_PRELIM_NO_DETECTION)
    cout << "Extraction : no preliminary detection (bs0 == NULL)." << endl;
  else if (res == BSDetector::RESULT_PRELIM_TOO_FEW)
    cout << "Extraction : two few points at preliminary detection." << endl;
  else if (res == BSDetector::RESULT_INITIAL_NO_DETECTION)
    cout << "Extraction : no initial detection (bsini == NULL)." << endl;
  else if (res == BSDetector::RESULT_INITIAL_TOO_FEW)
    cout << "Extraction : two few points at initial detection." << endl;
  else if (res == BSDetector::RESULT_INITIAL_TOO_SPARSE)
    cout << "Extraction : unsuccessful density test at initial detection."
         << endl;
  else if (res == BSDetector::RESULT_INITIAL_TOO_MANY_OUTLIERS)
    cout << "Extraction : unsuccessful filter test at initial detection."
         << endl;
  else if (res == BSDetector::RESULT_FINAL_NO_DETECTION)
    cout << "Extraction : no final detection (bsini == NULL)." << endl;
  else if (res == BSDetector::RESULT_FINAL_TOO_FEW)
    cout << "Extraction : two few points at final detection." << endl;
  else if (res == BSDetector::RESULT_FINAL_TOO_SPARSE)
    cout << "Extraction : unsuccessful density test at final detection."
         << endl;
  else if (res == BSDetector::RESULT_FINAL_TOO_MANY_OUTLIERS)
    cout << "Extraction : unsuccessful filter test at final detection."
         << endl;
}


void BSDetectionWidget::extract (bool updateAll)
{
  exam = -1;
  if (p1.equals (p2))
  {
    displayBackground ();
    return;
  }

  detector.detectSelection (p1, p2);
  displayDetectionResult (updateAll);
}


void BSDetectionWidget::extractAll ()
{
  udef = false;
  exam = -1;
  bool formerMultiMode = detector.isMultiSelection ();
  if (! formerMultiMode) detector.switchMultiSelection ();

  detector.detectAll ();
  vector<BlurredSegment *> mbs = detector.getBlurredSegments ();
  cout << mbs.size () << " blurred segments detected" << endl;
  displayDetectionResult (true);

  if (! formerMultiMode) detector.switchMultiSelection ();
}


void BSDetectionWidget::alternateTest ()
{
  if (p1.equals (p2))
  {
    cout << "Stroke undefined" << endl;
    return;
  }

  if (++alternate == 3) alternate = 0;
  if (alternate == 0)
  {
  }
  else if (alternate == 1)
  {
  }
  else if (alternate == 2)
  {
  }
} 



void BSDetectionWidget::performanceTest ()
{
  if (detector.isMultiSelection ())
  {
    cout << "Automatic extraction test" << endl;
    clock_t start = clock ();
    for (int i = 0; i < 100; i++) detector.detectAll ();
    double diff = (clock () - start) / (double) CLOCKS_PER_SEC;
    cout << "Test run : " << diff << endl;
    extract ();
  }
  else
  {
    if (p1.equals (p2))
    {
      cout << "Stroke undefined" << endl;
      return;
    }
    cout << "Run test" << endl;
    clock_t start = clock ();
    for (int i = 0; i < 1000; i++) detector.detect (p1, p2);
    double diff = (clock () - start) / (double) CLOCKS_PER_SEC;
    cout << "Test run : " << diff << endl;
    extract ();
  }
}


void BSDetectionWidget::localTest ()
{
  cout << "Run test" << endl;

/*
  p1 = Pt2i (212, 169);
  p2 = Pt2i (232, 152);
*/

  p1 = Pt2i (157, 125);
  p2 = Pt2i (165, 88);

  extract ();
  cout << "Test run" << endl;
}
